%%% The contents of this file are subject to the Erlang Public License,
%%% Version 1.0, (the "License"); you may not use this file except in
%%% compliance with the License. You may obtain a copy of the License at
%%% http://www.erlang.org/license/EPL1_0.txt
%%%
%%% Software distributed under the License is distributed on an "AS IS"
%%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%%% the License for the specific language governing rights and limitations
%%% under the License.
%%%
%%% The Original Code is xmerl-0.14
%%%
%%% The Initial Developer of the Original Code is Ericsson Telecom
%%% AB. Portions created by Ericsson are Copyright (C), 1998, Ericsson
%%% Telecom AB. All Rights Reserved.
%%%
%%% Contributor(s): ______________________________________.
%%%
%%%----------------------------------------------------------------------
%%% #0.    BASIC INFORMATION
%%%----------------------------------------------------------------------
%%% File:       xmerl.erl
%%% Author       : Ulf Wiger <ulf.wiger@ericsson.com>
%%% Description  : Functions to export simple and complete XML forms
%%% 
%%% Modules used : lists, <callbacks specified via arguments>
%%% 
%%%----------------------------------------------------------------------

-module(xmerl).

-compile(export_all).

-export([export/2, 
	 export/3,
	 export_simple/2, 
	 export_simple/3,
	 export_element/2,
	 export_element/3,   % special purposes only!
	 export_content/2,
	 callbacks/1]).

-include("xmerl.hrl").


%% export(Data, Callback) -> ExportedFormat.
%%
%% This function operates on "proper" XML records or content.
%%	- #xmlText{}
%%	- #xmlElement{}
%%	- #xmlPI{}
%%	- #xmlComment{}
%%	- #xmlDecl{}
%%	- Content
%%
%% Text can be deep lists of characters and/or binaries.
%%
export(Data, Callback) ->
    export(Data, Callback, []).

export(Data, Callback, RootAttrs) when atom(Callback) ->
    export1(Data, callbacks(Callback), RootAttrs);
export(Data, Callbacks, RootAttrs) when list(Callbacks) ->
    export1(Data, Callbacks, RootAttrs).

export1(Data, Callbacks, RootAttrs) ->
    Result = do_export(Data, Callbacks),
    Attrs = xmerl_lib:expand_attributes(RootAttrs, [{'#root#',1}]),
    Root = #xmlElement{name = '#root#',
		       pos = 1,
		       parents = [],
		       attributes = Attrs},
    Args = [Result, Root#xmlElement.attributes, [], Root],
    tagdef('#root#',1,[],Args,Callbacks).

%% This exports both elements and content.

do_export(Data, CBs) when list(Data) ->
    export_content(Data, CBs);
do_export(Data, CBs) ->
    export_element(Data, CBs).

%% export_simple(Data, Callback) -> ExportedFormat.
%%
%% This function operates on "simple" XML structures (this includes deep
%% strings), or content wrapped in a 1-tuple, as {Content}.
%%
%%	 - {Tag, Attr, Content}
%%	 - {Tag, Content}
%%	 - Tag
%%       - ProperXMLRecord
%%	 - DeepString
%%
export_simple(Data, Callback) ->
    export_simple(Data, Callback, []).

export_simple(Data, Callback, RootAttrs) when atom(Callback) ->
    export_simple1(Data, callbacks(Callback), RootAttrs);
export_simple(Data, Callbacks, RootAttrs) when list(Callbacks) ->
    export_simple1(Data, Callbacks, RootAttrs).

export_simple1(Data, Callback, RootAttrs) ->
    export(xmerl_lib:expand(Data), Callback, RootAttrs).

%% This exports flat, proper content lists. (We make a small
%% optimization here by skipping PI, Comment and Decl tags directly.)

export_content([E = #xmlText{} | T], CBs) ->
    [export_element(E, CBs) | export_content(T, CBs)];
export_content([E = #xmlElement{} | T], CBs) ->
    [export_element(E, CBs) | export_content(T, CBs)];
export_content([#xmlPI{} | T], CBs) ->
    export_content(T, CBs);
export_content([#xmlComment{} | T], CBs) ->
    export_content(T, CBs);
export_content([#xmlDecl{} | T], CBs) ->
    export_content(T, CBs);
export_content([], _CBs) ->
    [].

%% This is the normal element export.

export_element(#xmlText{value = Text}, CBs) ->
    apply_text_cb(CBs, Text);
export_element(E = #xmlElement{name = Tag,
			  pos = Pos,
			  attributes = Attributes,
			  parents = Parents,
			  content = Content}, CBs) ->
    Data = export_content(Content, CBs),
    Args = [Data, Attributes, Parents, E],
    tagdef(Tag,Pos,Parents,Args,CBs);
export_element(#xmlPI{}, _CBs) ->
    [];
export_element(#xmlComment{}, _CBs) ->
    [];
export_element(#xmlDecl{}, _CBs) ->
    [].

%% This is for special purposes only. Note that Content is passed to
%% the tag handler without first being exported recursively.

export_element(#xmlText{value = Text},CBs,_CBArgs) ->
    apply_text_cb(CBs, Text);
export_element(#xmlElement{name = Tag,
			   pos = Pos,
			   parents = Parents,
			   attributes = Attributes,
			   content = Content},CBs,CBArgs) ->
    Args = [Content, Attributes] ++ CBArgs,
    tagdef(Tag,Pos,Parents,Args,CBs);
export_element(#xmlPI{},_CBs,_CBArgs) ->
    [];
export_element(#xmlComment{},_CBs,_CBArgs) ->
    [];
export_element(#xmlDecl{},_CBs,_CBArgs) ->
    [].

%% A thing returned with #xml-redefine is assumed to be a single
%% element, unless it is a 1-tuple {Content} containing a content list.
%% The new element/content may be on "simple" format.

tagdef(Tag,Pos,Parents,Args,CBs) ->
    case apply_tag_cb(CBs, Tag, Args) of
	{'#xml-alias#', NewTag} ->
	    tagdef(NewTag,Pos,Parents,Args,CBs);
	{'#xml-redefine#', {NewContent}} ->
	    do_export(xmerl_lib:expand(NewContent, Pos, Parents), CBs);
	Other ->
	    Other
    end.

callbacks(M) ->
    Result = check_inheritance(M, []),
%%%     io:format("callbacks = ~p~n", [lists:reverse(Result)]),
    lists:reverse(Result).

callbacks([M|Mods], Visited) ->
    case lists:member(M, Visited) of
	false ->
	    NewVisited = check_inheritance(M, Visited),
	    callbacks(Mods, NewVisited);
	true ->
	    exit({cyclic_inheritance, {M, hd(Visited)}})
    end;
callbacks([], Visited) ->
    Visited.

check_inheritance(M, Visited) ->
%%%     io:format("calling ~p:'#xml-inheritance#'()~n", [M]),
    case M:'#xml-inheritance#'() of
	[] ->
	    [M|Visited];
	Mods ->
	    callbacks(Mods, [M|Visited])
    end.

apply_text_cb(Ms, Text) ->
    apply_cb(Ms, '#text#', '#text#', [Text]).

apply_tag_cb(Ms, F, Args) ->
    apply_cb(Ms, F, '#element#', Args).

apply_cb(Ms, F, Df, Args) ->
    apply_cb(Ms, F, Df, Args, Ms).

apply_cb([M|Ms], F, Df, Args, Ms0) ->
    case catch apply(M, F, Args) of
	{'EXIT', {undef, _}} ->
	    apply_cb(Ms, F, Df, Args, Ms0);
	{'EXIT', Reason} ->
	    exit(Reason);
	Res ->
	    Res
    end;
apply_cb([], Df, Df, Args, _Ms0) ->
    exit({unknown_tag, {Df, Args}});
apply_cb([], F, Df, Args, Ms0) ->
    apply_cb(Ms0, Df, Df, [F|Args]).
