%%% The contents of this file are subject to the Erlang Public License,
%%% Version 1.0, (the "License"); you may not use this file except in
%%% compliance with the License. You may obtain a copy of the License at
%%% http://www.erlang.org/license/EPL1_0.txt
%%%
%%% Software distributed under the License is distributed on an "AS IS"
%%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%%% the License for the specific language governing rights and limitations
%%% under the License.
%%%
%%% The Original Code is xmerl-0.6
%%%
%%% The Initial Developer of the Original Code is Ericsson Telecom
%%% AB. Portions created by Ericsson are Copyright (C), 1998, Ericsson
%%% Telecom AB. All Rights Reserved.
%%%
%%% Contributor(s): ______________________________________.
%%%
%%%----------------------------------------------------------------------
%%% #0.    BASIC INFORMATION
%%%----------------------------------------------------------------------
%%% File:       xmerl_lib.erl
%%% Author       : Ulf Wiger <ulf.wiger@ericsson.com>
%%% Description  : Utility module for handling XML trees.
%%%----------------------------------------------------------------------

-module(xmerl_lib).

-export([expand/1, expand/3, expand_content/1, expand_content/2,
	 expand_content/3, expand_element/1, expand_element/2,
	 expand_element/3, expand_attributes/1, expand_attributes/2,
	 expand_attributes/3, export_text/1, export_attribute/1,
	 markup/2, markup/3, simplify/1, simplify_element/1,
	 simplify_content/1, start_tag/1, start_tag/2, end_tag/1,
	 empty_tag/1, empty_tag/2,is_empty_data/1, find_attribute/2,
	 remove_whitespace/1]).

-include("xmerl.hrl").


%% Flatten text and escape special characters `<' and `&'.

export_text(T) ->
    export_text(T, []).

export_text([$< | T], Cont) ->
    "&lt;" ++ export_text(T, Cont);
export_text([$& | T], Cont) ->
    "&amp;" ++ export_text(T, Cont);
export_text([C | T], Cont) when integer(C) ->
    [C | export_text(T, Cont)];
export_text([T | T1], Cont) ->
    export_text(T, [T1 | Cont]);
export_text([], [T | Cont]) ->
    export_text(T, Cont);
export_text([], []) ->
    [];
export_text(Bin, Cont) ->
    export_text(binary_to_list(Bin), Cont).


%% Flatten attribute text and escape characters `"', `<' and `&'. (Note
%% that single-quote characters are not escaped; the markup-generating
%% functions (`start_tag', `end_tag', ...) always use `"' to delimit the
%% attribute values.)

export_attribute(T) ->
    export_attribute(T, []).

export_attribute([$< | T], Cont) ->
    "&lt;" ++ export_attribute(T, Cont);
export_attribute([$& | T], Cont) ->
    "&amp;" ++ export_attribute(T, Cont);
export_attribute([$" | T], Cont) ->
    "&quot;" ++ export_attribute(T, Cont);
export_attribute([C | T], Cont) when integer(C) ->
    [C | export_attribute(T, Cont)];
export_attribute([T | T1], Cont) ->
    export_attribute(T, [T1 | Cont]);
export_attribute([], [T | Cont]) ->
    export_attribute(T, Cont);
export_attribute([], []) ->
    [];
export_attribute(Bin, Cont) ->
    export_attribute(binary_to_list(Bin), Cont).


%% SimpleContent: [SimpleElement]
%% SimpleElement: #xml...{} | String | {atom(), [Attr], SimpleContent}
%%                | {atom(), SimpleContent} | atom()
%% Attr: {atom(), term()} | #xmlAttribute{}
%% String: [char() | binary() | String]
%%
%% Because strings can be deep, we do not allow content lists to also
%% be deep; otherwise, traversal of the simple representation becomes
%% too complicated and expensive. Simple content lists are thus flat
%% lists of simple elements.

%% The 'expand' function expands simple-form elements (this includes
%% deep strings), or content wrapped in a 1-tuple, as {Content}.

expand(Data) ->
    expand(Data, 1, []).

expand({Content}, Pos, Parents) ->
    expand_content(Content, Pos, Parents);
expand(Element, Pos, Parents) ->
    expand_element(Element, Pos, Parents).

%% Element expansion produces a single xml-record, and assumes that
%% all immediate sublists, if any, are strings.

expand_element(Element) ->
    expand_element(Element, []).

expand_element(Element, Parents) ->
    expand_element(Element, 1, Parents).

expand_element(E = #xmlElement{}, Pos, Parents) ->
    E#xmlElement{pos = Pos,
		 parents = Parents};
expand_element(E = #xmlText{}, Pos, Parents) ->
    E#xmlText{pos = Pos,
	      parents = Parents};
expand_element(E = #xmlPI{}, Pos, _Parents) ->
    E#xmlPI{pos = Pos};
expand_element(E = #xmlComment{}, Pos, Parents) ->
    E#xmlComment{pos = Pos,
		 parents = Parents};
expand_element(E = #xmlDecl{}, _Pos, _Parents) ->
    E#xmlDecl{};
expand_element({Tag, Attrs, Content}, Pos, Parents) when atom(Tag) ->
    NewParents = [{Tag, Pos} | Parents],
    FullAttrs = expand_attributes(Attrs, NewParents),
    #xmlElement{name = Tag,
		pos = Pos,
		parents = Parents,
		attributes = FullAttrs,
		content = expand_content(Content, NewParents)};
expand_element({Tag, Content}, Pos, Parents) when atom(Tag) ->
    NewParents = [{Tag, Pos}|Parents],
    #xmlElement{name = Tag,
		pos = Pos,
		parents = Parents,
		attributes = [],
		content = expand_content(Content, NewParents)};
expand_element(Tag, Pos, Parents) when atom(Tag) ->
    #xmlElement{name = Tag,
		pos = Pos,
		parents = Parents,
		attributes = [],
		content = []};
expand_element(String, Pos, Parents) when list(String) ->
    #xmlText{pos = Pos,
	     parents = Parents,
	     value = String}.

%% Content must be a flat list of elements.

expand_content(Content) ->
    expand_content(Content, []).

expand_content(Content, Parents) ->
    expand_content(Content, 1, Parents).

expand_content([H | T], Pos, Parents) ->
    [expand_element(H, Pos, Parents) | expand_content(T, Pos+1, Parents)];
expand_content([], _Pos, _Parents) ->
    [].

expand_attributes(Attrs) ->
    expand_attributes(Attrs, []).

expand_attributes(Attrs, Parents) ->
    expand_attributes(Attrs, 1, Parents).

expand_attributes([H = #xmlAttribute{} | T], Pos, Parents) ->
    [H#xmlAttribute{pos = Pos}
     | expand_attributes(T, Pos+1, Parents)];
expand_attributes([{K, V} | T], Pos, Parents) when atom(K) ->
    [#xmlAttribute{name = K,
		   pos = Pos,
		   parents = Parents,
		   value = V}
     | expand_attributes(T, Pos+1, Parents)];
expand_attributes([], _Pos, _Parents) ->
    [].

%% We want "simplification" to yield a normal form, so we always
%% generate three-tuples for elements; never {Tag, Content} or Tag.

%% The 'simplify' function works both on elements and content lists.

simplify(Data) when list(Data) ->
    simplify_content(Data);
simplify(Data) ->
    simplify_element(Data).

simplify_element(#xmlElement{name = Tag, attributes = Attrs,
			     content = Content}) when atom(Tag) ->
    {Tag, simplify_attributes(Attrs), simplify_content(Content)};
simplify_element(#xmlText{value = String}) ->
    String;
simplify_element(E = #xmlPI{}) ->
    E;
simplify_element(E = #xmlComment{}) ->
    E;
simplify_element(E = #xmlDecl{}) ->
    E;
simplify_element({Tag, Attrs, Content}) when atom(Tag) ->
    {Tag, simplify_attributes(Attrs), simplify_content(Content)};
simplify_element({Tag, Content}) when atom(Tag) ->
    {Tag, [], simplify_content(Content)};
simplify_element(Tag) when atom(Tag) ->
    {Tag, [], []}.

simplify_content([H | T]) ->
    [simplify(H) | simplify_content(T)];
simplify_content([]) ->
    [].

simplify_attributes([#xmlAttribute{name = K, value = V} | T])
  when atom(K) ->
    [{K, V} | simplify_attributes(T)];
simplify_attributes([H = {K, _} | T]) when atom(K) ->
    [H | simplify_attributes(T)];
simplify_attributes([]) ->
    [].

%% Looking up an attribute value

find_attribute(Name, Attrs) ->
    case lists:keysearch(Name, #xmlAttribute.name, Attrs) of
	{value, #xmlAttribute{value = V}} ->
	    {value, V};
	false ->
	    false
    end.


markup(Tag, Data) ->
    markup(Tag, [], Data).

markup(Tag, Attrs, []) ->
    empty_tag(Tag, Attrs);
markup(Tag, Attrs, Data) ->
    [start_tag(Tag, Attrs), Data, end_tag(Tag)].

start_tag(TagStr) ->
    start_tag(TagStr, []).

start_tag(Tag, Attrs) when atom(Tag) ->
    start_tag(atom_to_list(Tag), Attrs);
start_tag(TagStr, []) ->
    ["<", TagStr, ">"];
start_tag(TagStr, Attrs) ->
    ["<", TagStr, attributes(Attrs), ">"].

empty_tag(Tag) ->
    empty_tag(Tag, []).

empty_tag(Tag, Attrs) when atom(Tag) ->
    empty_tag(atom_to_list(Tag), Attrs);
empty_tag(TagStr, []) ->
    ["<", TagStr, "/>"];
empty_tag(TagStr, Attrs) ->
    ["<", TagStr, attributes(Attrs), "/>"].

end_tag(Tag) when atom(Tag) ->
    end_tag(atom_to_list(Tag));
end_tag(TagStr) ->
    ["</", TagStr, ">"].

attributes(Attrs) ->
    [attr_string(A) || A <- Attrs].

attr_string(#xmlAttribute{name = K, value = V}) ->
    [" ", atom_to_list(K), "=\"", value_string(V), "\""].

value_string(I) when integer(I) ->
    integer_to_list(I);
value_string(A) when atom(A) ->
    atom_to_list(A);
value_string(Term) ->
    case io_lib:deep_char_list(Term) of
	true ->
	    export_attribute(Term);
	false ->
	    io_lib:format("~p", [Term])
    end.


is_empty_data([]) ->
    true;
is_empty_data([X | Xs]) ->
    case is_empty_data(X) of
	false ->
	    false;
	true ->
	    is_empty_data(Xs)
    end;
is_empty_data(_) ->
    false.


%% Removing normalised whitespace-only text segments.

remove_whitespace([#xmlText{value = " "} | Data]) ->
    remove_whitespace(Data);
remove_whitespace([E = #xmlElement{content = Content} | Data]) ->
    [E#xmlElement{content = remove_whitespace(Content)}
     | remove_whitespace(Data)];
remove_whitespace([Other | Data]) ->
    [Other | remove_whitespace(Data)];
remove_whitespace([]) ->
    [].
